if (!('contains' in String.prototype)) {
	  String.prototype.contains = function(str, startIndex) {
	    return ''.indexOf.call(this, str, startIndex) !== -1;
	  };
	}

if (typeof String.prototype.startsWith != 'function') {
    String.prototype.startsWith = function(prefix) {
        return this.slice(0, prefix.length) == prefix;
    };
}
		 
if (typeof String.prototype.endsWith != 'function') {
    String.prototype.endsWith = function(suffix) {
        return this.slice(-suffix.length) == suffix;
    };
}


var ngp = require('./ecxngp');                                                                           


/*
  Because GUI/ECX conceptual simplification is the purpose of this module, it incorporates a fair amount of "business logic". This logic tends to receive a request from the NG GUI and, sub rosa,
  makes requests of ECX that were formerly made by the Sencha GUI itself. For example, the new GUI can register a provider without first having to create an ECX identity. This is because identity creation
  can be requested as part of provider registration.

  Thus, a fundamental aspect of the business logic is decomposing a request from the NG GUI into several SERIALIZED steps each of which typically involves talking to ECX. For example, provider registration
  may specify to use an extant identity or to create a new identity. The "serial function streams" defined below define an array of serial steps. For example, regProviderSerialFuncStream contains the
  3 steps required to accomplish both identity creation (if requested) and provider registration. Strictly speaking, each step defines the means of getting or creating the relevent data at ECX. The code
  in recursive function runSerialFuncs drives the actual processing of that data.

  As endpoints are added, each will likely require such an array.
*/

const urlParse = require('url-parse');

var funcSteps = [                                                                                 
	{                                                                                             
		continue: 'next',         
		name: 'reg_site',                                                                      
		verb: 'POST',    
		path: null,                                                                                
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              
		continue: 'next',                                                                          // veuser is->reg_usermode, else ->reg_vecreateuser
		name: 'reg_veusermode',
		verb: 'GET',
		path: '/api/identity/user',
		ip: null,                                                                                                                                       
		port: null, 
		type: null,
		func: null,                                                                                // see init 
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              
		continue: 'next',                                                                          // set by previous step [ next | skip | fail ]     
		name: 'reg_vecreateuser',                                                                  // -> reg_usermode
		verb: 'POST',
		path: '/api/identity/user',                                                            
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	},	
	{                                                                                              
		continue: 'next',                                                                          
		name: 'reg_usermode',                                                                      // hypervisor user is->reg_provider, else -> reg_createuser
		verb: 'GET',
		path: '/api/identity/user',
		ip: null,                                                                                                                                       
		port: null, 
		type: null,
		func: null,                                                                                 
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              
		continue: 'next',                                                                          // set by previous step [ next | skip | fail ]     
		name: 'reg_createuser',                                                                    // -> reg_provider
		verb: 'POST',
		path: '/api/identity/user',                                                            
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              
		continue: 'next',                                                                          // set by previous step [ next | skip | fail ]     
		name: 'reg_createuser_Instances',                                                                    // -> reg_provider
		verb: 'POST',
		path: '/api/identity/user',                                                            
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              
		continue: 'next',                                                                          // set by previous step [ next | skip | fail ]     
		name: 'reg_createkeys',                                                                    // -> reg_provider
		verb: 'POST',
		path: '/api/identity/key',                                                            
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              // NB: this block is changed before registration to conform to entity registering
		continue: 'next',         
		name: 'reg_provider',                                                                      // ng->reg_deleteuser; else ->reg_deleteveuser
		verb: 'POST',    
		path: null,                                                                                // NB: depends on type (hypervisor, storage, appserver,scriptserver)  
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
	{                                                                                              
		continue: 'next',                                                                          // set to skip if reg_provider succeeds
		name: 'reg_deleteuser',                                                                    // if driven delete any user identities that were created along the way
		verb: 'DELETE',    
		path: null,   
		ip: null, 
		port: null, 
		type: null,
		cb: function(res, data){ ngp.logMsg('info', 'Step=' + this.name + ' finished');}
	}, 
];

var HashMap = require('hashmap');

const regProviderStepFuncs = {
	reg_site: function(step, funcStream, ecxSessionId, oFromNg, cb){                                 //@@ Needs to know to set opProperties in the POST body
		
		regSite(step, funcStream, ecxSessionId, oFromNg, cb);		
	},
	reg_veusermode: function(step, funcStream, ecxSessionId, oFromNg, cb){
		if ( ngp.hasPropertyPath(oFromNg, 'opProperties.veServerInfo'.split('.') ) )                             // present only on hypervisor registration
		{
// It is incumbent upon the called routine to reset oFromNg.ve because these calls are asynchronous; thus we might unset it here before the call finishes or before .ve has been interrogated.			
			oFromNg.ve = true;                                                                           // tell common routine that we're interested in veuser for this call                                                         
			ngp.userMode(step, funcStream, ecxSessionId, oFromNg, cb);                                   //@@!! NB: before control ever returns here, the next step will be driven - LISTEN
		}
		else                                                                                             // simply recurse to next step reg_vecreateuser who will either create VE user or recurse to hypv user mode
		{
			step.cb(ecxSessionId, null);                                                                 
			ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	
		}
	},
	reg_vecreateuser: function(step, funcStream, ecxSessionId, oFromNg, cb){ 
		if ( ngp.hasPropertyPath(oFromNg, 'opProperties.veServerInfo'.split('.') ) ) 					
		{
			oFromNg.ve = true; 
			ngp.createUser(step, funcStream, ecxSessionId, oFromNg, cb);
}
		else
		{
			step.cb(ecxSessionId, null);                                                                 
			ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	
		}
	},	
	reg_usermode: function(step, funcStream, ecxSessionId, oFromNg, cb){                                 // NB: may have handled the VE user stuff, but must always handle provider user
		ngp.userMode(step, funcStream, ecxSessionId, oFromNg, cb);                                       // this call concerns the hypervisor user userMode
	},
	reg_createuser: function(step, funcStream, ecxSessionId, oFromNg, cb){                               
		ngp.createUser(step, funcStream, ecxSessionId, oFromNg, cb);                                     // create 'regular' (hypervisor or whatever) user as needed
	},
	reg_createuser_Instances: function(step, funcStream, ecxSessionId, oFromNg, cb) { 
		var instances = oFromNg.instances;
		var appCredentialsFromPostBody = [];
		 
		for (var i=0; i<instances.length; i++) {
			if( instances[i].user.username) {
					appCredentialsFromPostBody.push(instances[i]);
			}
		}
		   
		for (var i=0; i<appCredentialsFromPostBody.length; i++) {
			if(appCredentialsFromPostBody[i].user.username) {
				var idenityPostBody = {
					name: require('querystring').unescape(appCredentialsFromPostBody[i].user.username + '_' + oFromNg.hostAddress + '_' + appCredentialsFromPostBody[i].name),    // ****NB: can't create standard username_resourcename here****
					password: appCredentialsFromPostBody[i].user.password,
					username: require('querystring').unescape(appCredentialsFromPostBody[i].user.username),
				};
				ngp.createUserInstances(step, funcStream, ecxSessionId, oFromNg, idenityPostBody, i, appCredentialsFromPostBody.length, cb);   
			}
		}                                  
	},
	reg_createkeys: function(step, funcStream, ecxSessionId, oFromNg, cb) { 
		var instances = oFromNg.instances;
		ngp.createKey(step, funcStream, ecxSessionId, oFromNg, instances[0], cb);   
	},

	reg_provider: function(step, funcStream, ecxSessionId, oFromNg, cb){                                 //@@ Needs to know to set opProperties in the POST body
		if ( ngp.hasPropertyPath(oFromNg, 'opProperties.veServerInfo'.split('.') ) ) 
			oFromNg.ve = true;

		regProvider(step, funcStream, ecxSessionId, oFromNg, cb);		
	},
	reg_deleteuser: function(step, funcStream, ecxSessionId, oFromNg, cb){                               //@@ if this allowed to run, then MUST rollback any VE user too
		ngp.regDeleteUser(step, funcStream, ecxSessionId, oFromNg, cb);
	},
};


exports.init = init;


function init(req, res, providerType)                                                      // NB: hypervisor, storage, appserver, scriptserver
{
	var ngState = ngp.getSessionState(req, res);

	if ( ngState.stale )                                                                     // FTI we got bounced while client remained up. AND response already sent.
		return;                                                                                // !!! R U CERTAIN THAT RESPONSE SENT IN ALL CASES HERE?

	var ecxSid = ngState.ecxSid;
	res.ecxSid = ecxSid;

	var typology = ngp.getRegistrationTypology(req, providerType);                                       
    
/*
  We must clone the array of function steps so as to handle subsequent calls to this endpoint. But there are two cautions in cloning
    1. most clone methods cannot handle an object reference, i.e., an object in the clone source is preserved in the clone (same object is referenced by each)
	2. functions within a clone source object are lost and must be 're-attached'
*/

	var stepsStream = JSON.parse(JSON.stringify(funcSteps));                                    

	var regStep = null;

	stepsStream.forEach(function(step){                                                      // find the registration step
		if ( regStep )
			return;

		if ( step.name == 'reg_provider' )
		{
			regStep = step;
		}
	});

	regStep.path = typology.path;                                                            // NB: proper URL path against which to POST

	stepsStream.forEach(function(step){
		step.ip = ngp.localIp;
		step.port = ngp.ecxExternal;

		step.resourceType = typology.resourceType;
		step.type = typology.type;                                                             // NB: we need to be told the API path for type we are registering

		step.func = regProviderStepFuncs[step.name];                                           // hook up the step's real workhorse
		
		step.cb = function(ecxSid, data){ 
			//@#@ ngp.logMsg('info', "Finished step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'");
			ngp.logMsg('info', "Finished step '" + ngp.getCurrentFunc(step.reqId, ecxSid) + '(' + this.name + ")' in session ID '" + ecxSid + "'");

			if ( data )
			{
				ngp.logMsg('debug', "Data returned by step '" + ngState.currentFunc + '(' + this.name + ")' in session ID '" + ecxSid + "'=" + JSON.stringify(data));
			}
		};

		if ( step.name == 'reg_createuser' )                                                   // NB: failsafe function called only if reg_createuser fails
		{
			step.failsafe = function(step, funcStream, ecxSessionId, oFromNg, cb){
				var stepHash = ngp.arrayToHash(funcStream, 'name');
				funcStream.unshift(stepHash.reg_deleteuser);
				ngp.logMsg('warn', 'Function stream with failsafe step=' + JSON.stringify(funcStream));				
			};
		}

		if ( step.name == 'reg_createuser_Instances' )  {
			if(!req.body.instances) {
				step.continue = 'skip';
			} else if (req.body.instances[0].tenant) {
				step.continue = 'skip';
			} else if (!req.body.instances[0].tenant) {
				var instances = req.body.instances;
				var appCredentialsFromPostBody = [];
		 
				for (var i=0; i<instances.length; i++) {
					if( instances[i].user.username) {
						appCredentialsFromPostBody.push(instances[i]);
					}
				}

				if(appCredentialsFromPostBody.length == 0)
				step.continue = 'skip';
			}
		}	

		if ( step.name == 'reg_createkeys' )  {
			if(!req.body.instances)
			step.continue = 'skip';
			else if(!req.body.instances[0].tenant) {
				step.continue = 'skip';
			}
		}	
	});

	ngp.createRequestStore(stepsStream, ecxSid, req.reqId, 'register_' + typology.resourceType + '_' + typology.type);  

	ngState.currentFunc = 'register_' + typology.resourceType + '_' + typology.type;         // e.g., register_storage_netapp
	return stepsStream;
}


var regProvider = function(step, funcStream, ecxSessionId, oFromNg, cb)
{
	var scrubbedNgInput = ngp.scrubPostBody(oFromNg, true);
	ngp.logMsg('info', "Input from client for provider registration in session ID='" + ecxSessionId + "'=" + JSON.stringify(scrubbedNgInput));
	
	var ngState = ngp.ecxSidHash[ecxSessionId];
	var postBody = getProviderRegistrationPostBody(ecxSessionId, step, ngState, oFromNg);

	ngp.ngpCoreHelper.getSites(ecxSessionId, function(getSitesResObj) {
		ngp.ngpCoreHelper.getIdentities(ecxSessionId, function(getIdentitiesResObj) {
			ngp.ngpCoreHelper.getKeys(ecxSessionId, function(getKeysResObj) {
				if (oFromNg.siteName) {
					var arrayOfSites = getSitesResObj.rO['sites'];
		    		for (var i = 0; i < arrayOfSites.length; i++) {
						if (arrayOfSites[i]['name'] == oFromNg.siteName) {
							postBody.siteId = arrayOfSites[i]['id']
					   		break;
			 		   	}
		    		}
				}

				if((step.resourceType == "appserver") && (oFromNg.instances)) {
					if(oFromNg.applicationType != "office365") {
						var arrayOfIdentities = getIdentitiesResObj.rO['users'];
						var mapIdentitesNameId = new HashMap();
						
						for (var i = 0; i < arrayOfIdentities.length; i++) {
							mapIdentitesNameId.set(arrayOfIdentities[i]['name'],arrayOfIdentities[i]['links']['self']['href']);
						}
			
						var instances = oFromNg.instances;	
						var appCredentialsFromPostBody = [];
		
						for (var i=0; i<instances.length; i++) {
							if( ! instances[i].user.username) {
								appCredentialsFromPostBody.push({
									appuser : {
										href:instances[i].user
									},
									useForAllInstances:false,
									instanceName: instances[i].name
								});
							}
							else {
								appCredentialsFromPostBody.push({
									appuser : {
										href: mapIdentitesNameId.get(require('querystring').unescape(instances[i].user.username + '_' + oFromNg.hostAddress + '_' + instances[i].name))
									},
									useForAllInstances:false,
									instanceName: instances[i].name	
								});
							}
						}
			
						postBody.appCredentials = appCredentialsFromPostBody;
	
					} else {
						var arrayOfKeys = getKeysResObj.rO['keys'];
						var mapKeysNameId = new HashMap();
						
						for (var i = 0; i < arrayOfKeys.length; i++) {
							mapKeysNameId.set(arrayOfKeys[i]['name'],arrayOfKeys[i]['links']['self']['href']);
						}
			
						var instances = oFromNg.instances;	
						var appCredentialsFromPostBody = [];

						var keyDeleteHREF = mapKeysNameId.get(instances[0].name);
						var deleteKeyPath = urlParse(keyDeleteHREF).pathname;
		
						for (var i=0; i<instances.length; i++) {
							appCredentialsFromPostBody.push({
								appuser : {
									href: mapKeysNameId.get(instances[i].name)
								},
								useForAllInstances:false,
								instanceName: instances[i].tenant
							});
						}
			
						postBody.appCredentials = appCredentialsFromPostBody;
					}
				}

	    		ngp.sendRequestToEcx(step.verb, ngp.ecxExternal, step.path, ecxSessionId, postBody, function(regO) { 
	        		if ( regO.e || regO.response.statusCode != 201 ) { 		  
						   var failResult = ngp.getGenericEcxErrorResponse(regO, ecxSessionId, step); 
						   
						   if(oFromNg.applicationType == "office365")
						   var createdIdentities = ngp.getRequestState(step.reqId, ecxSessionId, 'createdIdentities');
						   else
						   var createdIdentities = ngp.getRequestState(step.reqId, ecxSessionId, 'createdKeys');
						   
			   
			   			if (oFromNg.siteName) {
				    		ngp.sendRequestToEcx('DELETE', ngp.ecxExternal, ngp.ngpConstants.SITE_API + '/' + postBody.siteId , ecxSessionId, null, function(deleteSiteResponse) {
								if(deleteSiteResponse.response.statusCode != 204) {
									var failResult = {
										statusCode: deleteSiteResponse.response.statusCode,
										response: {rc: 'Failed to delete Site ' + oFromNg.siteName}																
									};
														
						    		ngp.logMsg('warn', 'Failed to delete Site ' + oFromNg.siteName);
						
						    		if ( ! createdIdentities || createdIdentities.length < 1 ) {
										if ( funcStream.length > 0 )                                               
							    		funcStream[0].continue = 'skip'; 	
							    		ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, true);       				
						    		}
						    		else {
										ngp.setSerialStreamResult(ecxSessionId, failResult);
						        		step.cb(ecxSessionId, regO.rO);                                                  
						        		ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	      
						    		}			  			   
								} else {
									var failResult = ngp.getGenericEcxErrorResponse(regO, ecxSessionId, step);
						    		if ( ! createdIdentities || createdIdentities.length < 1 ) {
										if ( funcStream.length > 0 )                                              
							   			funcStream[0].continue = 'skip'; 	
							   			ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, true);     
						    		} else {
										ngp.setSerialStreamResult(ecxSessionId, failResult);                               
						       			step.cb(ecxSessionId, regO.rO);                                                  
						       			ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 	                    
						    		}  			   
					   			}
							});
						}
						else {
							if ( ! createdIdentities || createdIdentities.length < 1 ) {
								if ( funcStream.length > 0 )                                              
					    		funcStream[0].continue = 'skip'; 	
					    		ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, true);       // boolean arg causes ssfr not to set 'fail' (we've already set 'skip')				
				    		}
				    		else {
								if(oFromNg.applicationType == "office365") {
									ngp.sendRequestToEcx('DELETE', ngp.ecxExternal, deleteKeyPath, ecxSessionId, null, function(dlObj){  
										if ( dlObj.response.statusCode != 204 )
										ngp.logMsg('error', "Deletion of SPP Key " + deleteKeyPath + "' failed because of error=" + JSON.stringify(dlObj.rO));
										else
										ngp.logMsg('warn', "Deleted SPP Key - " + deleteKeyPath);
										
										ngp.setSerialStreamResult(ecxSessionId, failResult);                               
										step.cb(ecxSessionId, regO.rO);                                                  
										ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 
									});	        
								} else {
									ngp.setSerialStreamResult(ecxSessionId, failResult);                               
									step.cb(ecxSessionId, regO.rO);                                                  
									ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 
								}              
							}	  										   
			    		}			   
		    		}
					else  {                                                                                  // EHI provider registered				
			   			ngp.logMsg('debug', "Registration OK response=" + JSON.stringify(regO.rO));  
	
			    		var successResult = {
				  			statusCode: regO.response.statusCode,
				  			response: regO.rO,
			    		};

			    		ngp.setSerialStreamResult(ecxSessionId, successResult);                              // NB: let caller see result of successful update

			    		if ( funcStream.length > 0 ) {
							funcStream[0].continue = 'skip';                                                   // cause skip of next serial step (deletion of identities)                        
		        		}

		 	    		step.cb(ecxSessionId, regO.rO);                                                  
		    			ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 					
		    		}
				});
			});
		});
	});
};

var regSite = function(step, funcStream, ecxSessionId, oFromNg, cb)
{
	if(oFromNg.siteName)
	{
    	var siteCreationPostBody = null;
	
	    siteCreationPostBody = {  
	      name: oFromNg.siteName,              
	      description: '',                                                             
	      defaultSite: false, 
	    };			   

	    ngp.sendRequestToEcx('POST', ngp.ecxExternal, ngp.ngpConstants.SITE_API, ecxSessionId, siteCreationPostBody, function(regO){ 
		    if ( regO.e || regO.response.statusCode != 201 )   
		    {
			  ngp.logMsg('info', "Site Creation Failed response=" + JSON.stringify(regO.rO));
			  var failResult = ngp.getGenericEcxErrorResponse(regO, ecxSessionId, step); 
			  ngp.setStepFailResult(step, funcStream, ecxSessionId, oFromNg, cb, failResult, false); 
		    }
		    else                                                                                   
		    {
			  ngp.logMsg('info', "Site successfully created OK response=" + JSON.stringify(regO.rO));  
			  step.cb(ecxSessionId, regO.rO);                                                  
			  ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 
		    }		
	    });	
	}
	else
	ngp.runSerialFuncs(funcStream, ecxSessionId, oFromNg, cb); 
	
};

// If we are registering a hypervisor and oFromNg contains opProperties (used to establish VE server info, including an ECX identity), then we must return
// a POST body that includes the resolved opProperties.veServerInfo.username

function getProviderRegistrationPostBody(ecxSessionId, step, ngState, oFromNg)             // called by regProvider 
{
	var postBody = null;

	postBody = {  
		addToCatJob: true,              
		hostAddress: oFromNg.hostAddress,   
		maxStreams: oFromNg.maxStreams,                                                          
		portNumber: oFromNg.portNumber,                                                        // use whatever GUI gives us, if anything; else backend decides
		sslConnection: oFromNg.sslConnection,                                                  // ditto
		user: {
			//@#@href: ngp.getEndpointState(ecxSessionId, 'userHref'),                                // always correct to convey identity href for hypervisor user
			href: ngp.getRequestState(step.reqId, ecxSessionId, 'userHref'),                                // always correct to convey identity href for hypervisor user
		},
	};

	if ( oFromNg.ve )                                                                        // FTI must also provide href for opProperties.veServerInfo.username (extant | created)
	{
		oFromNg.opProperties.veServerInfo.osuser = {			
			//@#@href: ngp.getEndpointState(ecxSessionId, 'veUserHref'),                              //@@ EXTANT OR CREATED
			href: ngp.getRequestState(step.reqId, ecxSessionId, 'veUserHref'),                               //@@ EXTANT OR CREATED			
		};

		delete oFromNg.opProperties.veServerInfo.username;
		delete oFromNg.opProperties.veServerInfo.password;		
		postBody.opProperties = oFromNg.opProperties;
		delete oFromNg.ve;                                                                     // ESSENTIAL so as not to screw up remaining steps
	}

// Handle DIFFERENTIAE here. Note that funcO.resourceType is assigned by regProviderCommon based on the typology. It does NOT come from GUI.

	if ( step.resourceType == 'hypervisor' )
	{
		delete postBody.type;
		postBody.addToCatJob = true; 
		postBody.type = oFromNg.type;
		postBody.name = oFromNg.hostAddress; 
		postBody.opProperties = oFromNg.opProperties;                                          // NB should be safe even if already set by .ve case in if block above
	}

	if ( step.resourceType == 'storage' )
	{
		var hrefUrl = postBody.user.href;
		postBody.user = {href: hrefUrl};                                                       // SOMEDAY? Today, actually, July 18th, 2017...
		postBody.type = oFromNg.type;
		postBody.siteId = oFromNg.siteId;
		postBody.name = oFromNg.hostAddress; 
	}

//!! only HANA requires appCredentials: oFromNg.appCredentials,
	if (step.resourceType == 'appserver' )
	{
        postBody.addToCatJob = oFromNg.addToCatJob === undefined ? true : oFromNg.addToCatJob;
        postBody.application = oFromNg.application === undefined ? false : oFromNg.application;
		postBody.script = oFromNg.script === undefined ? true : oFromNg.script;

		var hrefPathName = urlParse(postBody.user.href).pathname;
		if(hrefPathName.includes('/api/identity/key')) 
		postBody.oskey = postBody.user;
		else
		postBody.osuser = postBody.user;

		delete postBody.user;

		postBody.osType = oFromNg.osType;
		postBody.applicationType = oFromNg.applicationType;
		postBody.name = oFromNg.hostAddress;  
		postBody.useKeyAuthentication = oFromNg.useKeyAuthentication;
		postBody.opProperties = oFromNg.opProperties;
		
		if(!oFromNg.portNumber)
		delete postBody.portNumber;
		
		delete postBody.sslConnection;		
	}

	if (step.resourceType == 'scriptserver' )
	{
        postBody.osuser = postBody.user;
		
		delete postBody.user;

		postBody.osType = oFromNg.osType;
		postBody.name = oFromNg.hostAddress;  
		postBody.useKeyAuthentication = oFromNg.useKeyAuthentication;

		delete postBody.portNumber;
		delete postBody.sslConnection;
		delete postBody.addToCatJob;
	}

	return postBody;	
}
